## three dimensional simulation of the three-body problem
## using the Python Visual module for the animation
## see http://www.vpython.org/
## this version is hacked to show Kepler's second law
from visual import *

global threebody
scene = display(title="University of Glasgow A1 Dynamical Astronomy. Kepler's Second Law (press 1-2). RH mouse rotates, LH+RH zooms.", width=1280, height=800)
scene.autoscale=false;

## set up the masses. m1 is missing in 2-body systems
m1 = sphere(color=color.red,   make_trail=True, interval=2, retain=1000)
m2 = sphere(color=color.white,make_trail=True, interval=2, retain=1000, material=materials.BlueMarble)
m3 = sphere(color=color.cyan, make_trail=True, interval=2, retain=1000)
kepler = curve()
scene.lights=[]
decsun = 20/180*pi
lamp = distant_light(direction = (cos(decsun)/sqrt(2),-cos(decsun)/sqrt(2),sin(decsun)),color=color.white)# the Sun221!

## move to the centre of mass frame and adjust sizes for constant density
def resetscene(m1,m2,m3):
    vcentre=(m1.mass*m1.vel+m2.mass*m2.vel+m3.mass*m3.vel)/(m1.mass+m2.mass+m3.mass)
    rcentre=(m1.mass*m1.pos+m2.mass*m2.pos+m3.mass*m3.pos)/(m1.mass+m2.mass+m3.mass)
    for a in [m1, m2, m3]:
        a.vel -= vcentre
        a.radius = 0.5*a.mass**(1.0/3.0)
        a.trail_object.pos=[]
    kepler.pos=[]

def dydt(y):
## determine the derivative of the state vector y
    deriv = zeros((6,3), dtype=vector)
    r12=y[0]-y[2]; r23=y[2]-y[4]; r31=y[4]-y[0]
    r12c=r12/mag(r12)**3; r23c=r23/mag(r23)**3; r31c=r31/mag(r31)**3
    deriv[1] = -m2.mass*r12c + m3.mass*r31c
    deriv[3] = -m3.mass*r23c + m1.mass*r12c
    deriv[5] = -m1.mass*r31c + m2.mass*r23c
    deriv[0:5:2] = y[1:6:2] ## just copy the three velocities from y
    return deriv


def keyInput(evt):
    global threebody,dt
    s = evt.key
    if (s == '1'): ##circular
        m1.pos=vector( 100,0,0); m1.vel=vector(0,0,0);   m1.mass=0
        m2.pos=vector( 0,0,0);   m2.vel=vector( 0, 0,0); m2.mass=10
        m3.pos=vector( 10,0,0);  m3.vel=vector(0,1,0);  m3.mass=0.01
        scene.range=20
        resetscene(m1,m2,m3)
        threebody = False;
    if (s == '2'):
        m1.pos=vector( 100,0,0); m1.vel=vector(0,0,0);   m1.mass=0
        m2.pos=vector( 5,0,0);   m2.vel=vector(0,0,0);   m2.mass=10.0
        m3.pos=vector(-10,0,0);  m3.vel=vector(0,-0.5,0); m3.mass=0.01
        scene.range=15
        resetscene(m1,m2,m3)
        threebody= False


scene.bind('keydown', keyInput)

## initiate with the simplest system
m1.pos=vector( 100,0,0); m1.vel=vector(0,0,0);   m1.mass=0
m2.pos=vector( 0,0,0);   m2.vel=vector( 0, 0,0); m2.mass=10
m3.pos=vector( 10,0,0);  m3.vel=vector(0,1,0);  m3.mass=0.01
scene.range=20
resetscene(m1,m2,m3)
m2.rotate(angle = pi/2,axis=(1,0,0),origin=m2.pos)
threebody = False;

dt = 0.1 ## the time step.  Keep this as small as possible.
c = 0
while True:
  rate(100)
  ## solve using forth-order Runge Kutta.  See CHOPF p62.
  y = [m1.pos,m1.vel,m2.pos,m2.vel,m3.pos,m3.vel]
  k1 = dt*dydt(y);  k2 = dt*dydt(y+k1/2.0)
  k3 = dt*dydt(y+k2/2.0);  k4 = dt*dydt(y+k3)
  dy = k1/6.0 + k2/3.0 +k3/3.0 + k4/6.0
## update the animation
  if threebody:
    m1.pos += dy[0]; m1.vel += dy[1]
  m2.pos += dy[2]; m2.vel += dy[3]
  m3.pos += dy[4]; m3.vel += dy[5]
  m2.rotate(angle = 0.0003,axis=(0,0,1),origin=m2.pos)
#  m2.rotate(angle = dt*0.1,axis=(0,0,1),origin=m2.pos)
  c+=1
  if (c % 30) == 0:
    kepler.append((m1.mass*m1.pos+m2.mass*m2.pos+m3.mass*m3.pos)/(m1.mass+m2.mass+m3.mass))
    kepler.append(pos=m3.pos,retain=12)
